/* floodfill.c
 * This file is part of Vara digital painting program
 * Copyright 2021, 2023, 2024 Nandakumar Edamana
 * Distributed under GNU General Public License, version 3
 */ 

#include <assert.h>
#include <stdarg.h>
#include <stdbool.h>

#include "floodfill.h"

void point_stack_construct(PointStack *this)
{
	this->count = 0;
	Point *_tmp_1 = (Point *) malloc(8 * sizeof(Point));
	if(_tmp_1 == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	this->arr = _tmp_1;
	this->alloccount = 8;
}

_Bool point_stack_is_empty(PointStack *this)
{
	return 0 == this->count;
}

void point_stack_push(PointStack *this, Point newitem)
{
	point_stack__resize(this, this->count + 1);
	this->arr[this->count - 1] = newitem;
}

void point_stack_push_multiple(PointStack *this, int n, Point *newitems)
{
	int i;
	int newcount = this->count + n;
	if(this->alloccount < newcount) {
		this->arr = realloc(this->arr, newcount * sizeof(this->arr[0]));
		if(this->arr == NULL) { perror(NULL); exit(EXIT_FAILURE); }

		this->alloccount = newcount;
	}

	for(i = 0; i < n; i += 1) {
		this->arr[this->count + i] = newitems[i];
	}

	this->count = newcount;
}

Point point_stack_peek(PointStack *this)
{
	assert(this->count > 0); /* stack.ngg:42 */

	return this->arr[this->count - 1];
}

Point point_stack_pop(PointStack *this)
{
	assert(this->count > 0); /* stack.ngg:48 */
	Point retitem = this->arr[this->count - 1];
	point_stack__resize(this, this->count - 1);

	return retitem;
}

Point point_stack_pop_bottom(PointStack *this)
{
	assert(this->count > 0); /* stack.ngg:61 */
	Point retval = this->arr[0];
	this->count -= 1;
	memmove(this->arr, this->arr + 1, this->count * sizeof(this->arr[0]));

	return retval;
}

void point_stack_replace_top(PointStack *this, Point newitem)
{
	assert(this->count > 0); /* stack.ngg:73 */
	this->arr[this->count - 1] = newitem;
}

void point_stack__resize(PointStack *this, int newcount)
{
	int UNIT = 8;
	int newalloccount = this->alloccount;
	if(this->alloccount != newcount) {
		if(this->alloccount < newcount) {
			newalloccount = this->alloccount + UNIT;
		} else {
			int spare = this->alloccount - newcount;
			if(spare > (2 * UNIT)) {
				newalloccount = this->alloccount - UNIT;
			}

		}

		assert(newalloccount >= this->count); /* stack.ngg:104 */
		if(newalloccount != this->alloccount) {
			if(newalloccount == 0) {
			} else {
				this->arr = realloc(this->arr, newalloccount * sizeof(this->arr[0]));
				if(this->arr == NULL) { perror(NULL); exit(EXIT_FAILURE); }

			}

			this->alloccount = newalloccount;
		}

	}

	this->count = newcount;
}

void point_stack_destruct(PointStack *this)
{
	if(this->arr) {
		free(this->arr);
	}

}

void boolean_map_construct(BooleanMap *this, int w, int h)
{
	Image *_tmp_1 = (Image *) malloc(sizeof(Image));
	if(_tmp_1 == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	image_construct(_tmp_1, w, h, 255);
	this->img = _tmp_1;
}

_Bool boolean_map_get(BooleanMap *this, int x, int y)
{
	return nan_img_get_pixel(this->img, x, y).rgb[0] != 0;
}

_Bool boolean_map_set(BooleanMap *this, int x, int y)
{
	nan_img_set_pixel(this->img, x, y, PIXEL_WHITE);
}

void boolean_map_destruct(BooleanMap *this)
{
	if(this->img) {
		image_destruct(this->img);
		free(this->img);
	}

}

struct Rectangle nan_img_floodfill(Image *img, int x, int y, struct DrawingParams dp)
{
	BooleanMap *donemap = (BooleanMap *) malloc(sizeof(BooleanMap));
	if(donemap == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	boolean_map_construct(donemap, img->width, img->height);
	PointStack *ptstack = (PointStack *) malloc(sizeof(PointStack));
	if(ptstack == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	point_stack_construct(ptstack);
	struct Pixel fillpxl = rgbpixel_with_alpha(dp.fgcolor, 1);
	struct Pixel ifpxl = nan_img_get_pixel(img, x, y);
	floodfill_add_point_to_stack_if_match(img, ptstack, donemap, (Point) { x, y }, ifpxl);
	int xmin = x;
	int xmax = x;
	int ymin = y;
	int ymax = y;
	while(!point_stack_is_empty(ptstack)) {
		Point point = point_stack_pop(ptstack);
		nan_img_set_pixel(img, point.x, point.y, fillpxl);
		if(point.x < xmin) {
			xmin = point.x;
		} else if(point.x > xmax) {
			xmax = point.x;
		}

		if(point.y < ymin) {
			ymin = point.y;
		} else if(point.y > ymax) {
			ymax = point.y;
		}

		if(point.x > 0) {
			floodfill_add_point_to_stack_if_match(img, ptstack, donemap, (Point) { point.x - 1, point.y }, ifpxl);
		}

		if(point.x < (img->width - 1)) {
			floodfill_add_point_to_stack_if_match(img, ptstack, donemap, (Point) { 1 + point.x, point.y }, ifpxl);
		}

		if(point.y > 0) {
			floodfill_add_point_to_stack_if_match(img, ptstack, donemap, (Point) { point.x, point.y - 1 }, ifpxl);
		}

		if(point.y < (img->height - 1)) {
			floodfill_add_point_to_stack_if_match(img, ptstack, donemap, (Point) { point.x, 1 + point.y }, ifpxl);
		}

	}

	if(donemap) {
		boolean_map_destruct(donemap);
		free(donemap);
	}

	if(ptstack) {
		point_stack_destruct(ptstack);
		free(ptstack);
	}


	return rectangle_new_from_boundaries(xmin, ymin, xmax, ymax);
}

void floodfill_add_point_to_stack_if_match(Image *img, PointStack *ptstack, BooleanMap *donemap, Point point, struct Pixel ifpxl)
{
	if(boolean_map_get(donemap, point.x, point.y)) {
		return ;
	}

	if(!pixelsmatch(ifpxl, nan_img_get_pixel(img, point.x, point.y))) {
		return ;
	}

	point_stack_push(ptstack, point);
}

Point point_default()
{
	Point s;
	s.x = 0;
	s.y = 0;

	return s;
}

